/*
 * symbol.c
 *
 * Copyright 1999 Michael Elizabeth Chastain, <mailto:mec@shout.net>.
 * Licensed under the Gnu Public License, Version 2.
 *
 * Symbol table.
 */

#include <string.h>

#include "mconfig.h"



/*
 * Some convenient piece constants.
 */

const piece_type piece_empty = { "",  0 };
const piece_type piece_y     = { "y", 1 };
const piece_type piece_m     = { "m", 1 };
const piece_type piece_n     = { "n", 1 };



/*
 * A little hash table.
 */

#define HASH_SIZE	1009

static symbol_type * symbol_table [HASH_SIZE];



/*
 * Compute a hash code.  This likes to hash on the last characters
 * of the string, which is fine, because the first characters are
 * almost always "CONFIG_" anyways.
 */

static unsigned int hash_key( piece_type name )
{
    int i;
    unsigned int n = 0;

    for ( i = 0; i < name.len; i++ )
	n = (n << 1) ^ (unsigned char) name.ptr [i];

    return n % HASH_SIZE;
}



/*
 * Lookup a symbol from a name.
 * Returns existing symbol if one is already available.
 * Creates new symbol if no symbol exists with that name.
 */

symbol_type * symbol_lookup( piece_type name )
{
    const unsigned int key = hash_key( name );
    symbol_type * symbol;

    for ( symbol = symbol_table [key]; symbol != NULL; symbol = symbol->next_hash )
    {
	if ( name.len == symbol->name.len
	&&   memcmp( name.ptr, symbol->name.ptr, name.len ) == 0 )
	    return symbol;
    }

    symbol             = grab_memory( sizeof(*symbol) );
    symbol->next_list  = NULL;
    symbol->name       = name;
    symbol->value      = piece_empty;
    symbol->origin     = origin_unset;
    symbol->next_hash  = symbol_table [key];
    symbol_table [key] = symbol;
    return symbol;
}



/*
 * Reset the value of all symbols to "undefined".
 * Also, populate the table with symbols $ARCH and $arch.
 */

void symbol_table_reset( )
{
    int i;
    symbol_type * symbol;

    for ( i = 0; i < HASH_SIZE; i++ )
    {
	for ( symbol = symbol_table[i]; symbol != NULL; symbol = symbol->next_hash )
	{
	    symbol->value  = piece_empty;
	    symbol->origin = origin_unset;
	}
    }

    if ( argument.arch != NULL )
    {
	piece_type name;
	piece_type value;

	name.ptr       = "ARCH";
	name.len       = 4;
	value.ptr      = argument.arch;
	value.len      = strlen(argument.arch);

	symbol         = symbol_lookup( name );
	symbol->value  = value;
	symbol->origin = origin_permanent;

	name.ptr       = "arch";
	symbol         = symbol_lookup( name );
	symbol->value  = value;
	symbol->origin = origin_permanent;
    }
}
